import math
import numpy as np
import sys
import statsmodels.api as sm
from scipy import stats
from scipy import spatial as sps
import os

# PointPattern attributes:
	# npoints: number of points in pattern
	# marks: array of point marks for each point, numel(marks) = npoints
	# umarks: list of all unique marks
	# marked: bool, True if points have marks, False if not 
	# multimarked: bool, True if multitype point pattern (treated as unmarked if False)
	# points: npoints x d array of cartesian coordinates of points where d is the dimension of the point pattern
	# cellpat: nc x d array of cartesian coordinates of cell membrane points, to ensure a well behaved shape convex hull of points is used, default is empty
	# nucpat: nn x d array of cartesian coordinates of nuclear membrane points, to ensure a well behaved shape convex hull of points is used, default is empty
	# box: d x 2 array of bounding box, [xmin xmax; ymin ymax; zmin zmax], if unset, default is minimum bounding box of points
	# cellarea: (may be removed), area of cellpat

# QuadScheme attributes:
	# pat: stores original point pattern for future reference
	# points: array of PointPattern points and dummy points (the first pat.npoints are observed points, the next quad.numdummy points are dummy points)
	# quadweights: npoints+numdummy x 1 array of weights defined for each point
	# responses: npoints+numdummy x 1 array of responses per the Berman Turner method
	# ptlabs: npoints+numdummy x 1 bool array where 0 = dummy points, 1 = observed point
	# ptsonedge: points, if any, that fall on an edge of a tile used in quadrature
	# nrp: number of observed points
	# numdummy: number of dummy points

#############################################################
# Classes

## this is used to check if optional arguments are supplied
# DEFAULT = object()
class DEFAULT:
	def __init__(self):
		self.self = 'default'

class PointPattern:
	"""

	Point pattern object, holds all necessary components of a point pattern.

	dimension : dimension of point pattern (int)
	points : array of point coordinates (nparray)
	npoints : number of observed points (int)
	marked : flag indicating whether a pattern is marked (bool)
	multimarked : flag indicating whether a pattern has multiple marks (bool)
	marks : a 1 by npoints list of strings, tuples, ints, or floats corresponding to the mark/label of each point. ORDER MATTERS HERE (list)
	umarks : unique marks in the point pattern (list)
	cellpat : array of cell membrane coordinates (nparray)
	cellflag : flag, true if cellpat is provided (bool)
	cellarea : area contained within cell (float)
	nucpat : array of nuclear membrane coordinates (nparray)
	nucflag : flag, true if nucpat is provided (bool)
	box : minimum box containing all points (np.array, [xmin, xmax; ymin, ymax; zmin, zmax])


	"""

	## main function to create a point pattern from data
	def __init__(self,pos,mark=DEFAULT(),area=DEFAULT(),cellpat=DEFAULT(),nucpat=DEFAULT(),box=DEFAULT(),dimension=DEFAULT,otherpat=DEFAULT):
		## get correct dimension
		if not isinstance(dimension,DEFAULT):
			self.dimension = dimension
		else:
			if np.shape(pos)[1] != dimension:
				print 'Specified dimension does not match point dimension, reverting to point dimension'
				self.dimension = np.shape(pos)[1]
			else:
				self.dimension = dimension

		## store the number of points in the pattern
		self.npoints = np.shape(pos)[0]

		if not isinstance(otherpat,DEFAULT):
			self.otherpat = otherpat
		## check if we have a marked point pattern
		if not isinstance(mark,DEFAULT):

			## if given a list, one mark for each point, use the list as our marks
			if len(mark) == self.npoints:
				self.marks = np.array(mark)
				self.umarks = list(set(mark))

			## otherwise, create a list, one mark for each point, all the same mark
			else:
				self.marks = np.tile(mark,self.npoints)
				self.umarks = [mark]
			
			## set the marked flag to true
			self.marked = True

			## check if we have multiple unique marks, assign multimarked flag accordingly
			if len(self.umarks)>1:
				self.multimarked = True
			else:
				self.multimarked = False
		else:
			self.marked = False
			self.multimarked = False
		
		## sort the points by x values in ascending order (this will speed calculations of nearest neighbors, covar dist, etc later)
		self.points = pos[pos[:,0].argsort()]

		## make sure that we put the marks in the correct sorted order too
		if self.marked:
			self.marks = [x for (y,x) in sorted(zip(pos[:,0],self.marks), key=lambda pair: pair[0])]
			self.marks = np.array(self.marks)

		## assign cellpat
		ctup = tuple(map(tuple, cellpat))
		ptup = tuple(map(tuple, pos))

		carr = np.array([x for x in ctup if x not in ptup])
		self.cellpat = sps.cKDTree(carr)

		## determine box from cellpat
		cxmin = np.amin(cellpat[:,0])
		cxmax = np.amax(cellpat[:,0])
		cymin = np.amin(cellpat[:,1])
		cymax = np.amax(cellpat[:,1])

		if self.dimension == 3:
			czmin = np.amin(cellpat[:,2])
			czmax = np.amax(cellpat[:,2])
			self.box = np.array([[cxmin,cxmax],[cymin,cymax],[czmin,czmax]])

		else:
			self.box = np.array([[cxmin,cxmax],[cymin,cymax]])

		self.nucflag = False
		nflag = False

		try:
			if isinstance(nucpat,DEFAULT):
				self.nucpat = []
				self.nucflag = False
				nflag = True
		except Exception, e:
			pass

		if not nflag:
			try:
				if not isinstance(nucpat,DEFAULT):
					self.nucpat = sps.cKDTree(nucpat)
					# self.nucpat = nucpat
			except Exception, e:
				self.nucpat = []
				print "Nuclear pattern in unrecognized format, skipping."

		## assign the cell area if input
		if not isinstance(cellpat,DEFAULT):
			if dimension == 3:
				nh = sps.ConvexHull(nucpat)
				ch = sps.ConvexHull(cellpat)
				self.cellarea = hullVol(ch)-hullVol(nh)
			else:
				nh = sps.ConvexHull(nucpat)
				ch = sps.ConvexHull(cellpat)
				self.cellarea = hullVol(ch)-hullVol(nh)
		else:
			self.cellarea = 0
		
		if self.cellarea == 0:
			if self.dimension == 2:
				self.cellarea = (self.box[0,1]-self.box[0,0])*(self.box[1,1]-self.box[1,0])
			elif self.dimension == 3:
				self.cellarea = (self.box[0,1]-self.box[0,0])*(self.box[1,1]-self.box[1,0])*(self.box[2,1]-self.box[2,0])

	## this allows you to add points to an already existing point pattern
	def addpoints(self,pos,mark=DEFAULT):
		self.npoints += np.shape(pos)[0]
		self.points = np.vstack((self.points,pos))

		if not isinstance(mark,DEFAULT):
			if not self.marked:
				print "Adding marked points to unmarked pattern, assigning arbitrary mark to original points."
				self.marked = True
				self.multimarked = True
				self.marks = ["." for x in xrange(0,self.npoints)]
				self.umarks = ["."]

			## if given a list, one mark for each point, use the list as our marks
			if len(mark) == np.shape(pos)[0]:
				self.marks = np.hstack((self.marks,np.array(mark)))
				self.umarks = list(set(self.marks))

			## otherwise, create a list, one mark for each point, all the same mark
			else:
				marks = np.tile(mark,np.shape(pos)[0])
				self.marks = np.hstack((self.marks,marks))
				self.umarks = list(set(self.marks))

			## check if we have multiple unique marks, assign multimarked flag accordingly
			if len(self.umarks)>1:
				self.multimarked = True
			else:
				self.multimarked = False
		else:
			if self.marked:
				print "Adding unmarked points to marked point pattern, assigning arbitrary mark to new points."
				self.multimarked = True
				marks = ["." for x in xrange(0,np.shape(pos)[0])]
				self.marks += marks
				self.umarks = list(set(self.marks))

		## sort the points by x values in ascending order (this will speed calculations of nearest neighbors, covar dist, etc later)
		self.points = pos[pos[:,0].argsort()]

		## make sure that we put the marks in the correct sorted order too
		if self.marked:
			self.marks = [x for (y,x) in sorted(zip(pos[:,0],self.marks), key=lambda pair: pair[0])]

	## prints all point coordinates and the bounding box to the console
	def printPat(self):
		for pt in xrange(0,self.npoints):
			print self.points[pt,:],self.marks[pt]
		print self.box

	## prints a relevant summary of the point pattern to the console
	def printSummary(self):
		if self.marked and not self.multimarked:
			sys.stdout.write("Marked "+str(self.dimension)+"D Point Pattern with "+str(self.npoints)+" points.\n")
			sys.stdout.write("Unique marks: ")
			for m in self.umarks:
				sys.stdout.write(str(m)+" ")
			sys.stdout.write("\n")
		elif self.multimarked:
			sys.stdout.write("Multimarked "+str(self.dimension)+"D Point Pattern with "+str(self.npoints)+" points.\n")
			sys.stdout.write("Unique marks: ")
			for m in self.umarks:
				sys.stdout.write(str(m)+" ")
			sys.stdout.write("\n")
		else:
			sys.stdout.write("Unmarked "+str(self.dimension)+"D Point Pattern with "+str(self.npoints)+" points.\n")

		sys.stdout.write("Attributes: ")
		if self.cellflag and self.nucflag:
			sys.stdout.write("cell pattern, nuclear pattern\n")
		elif self.cellflag:
			sys.stdout.write("cell pattern\n")
		elif self.nucflag:
			sys.stdout.write("nuclear pattern\n")
		else:
			sys.stdout.write("none\n")
		sys.stdout.write("\n")
			
	## plots the point pattern using matplotlib
	def plotPat(self):
		filled_markers = ('o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd')
		mcolors = ('b','g','r','c','m','y','k','w')
		if np.shape(self.points)[1] <= 2:
			#fig = plt.figure()
			#ax = fig.subplot()
			fig,ax = plt.subplots(1,1)
			for lab in self.umarks:
				points = self.points[self.marks == lab]
				curmarker = filled_markers[self.umarks.index(lab)]
				ax.scatter(points[:,0],points[:,1],c=mcolors[self.umarks.index(lab)],marker=curmarker)
			xmax, ymax = self.points.max(axis=0)
			xmin, ymin = self.points.min(axis=0)
			ax.set_xlim(xmin,xmax)
			ax.set_ylim(ymin,ymax)
			ax.set_xticklabels([])
			ax.set_yticklabels([])
			plt.show()
			
		elif np.shape(self.points)[1] == 3:
			fig = plt.figure()
			#ax = fig.add_subplot(111, projection='3d')
			ax = Axes3D(fig)
			for lab in self.umarks:
				points = self.points[self.marks == lab]
				curmarker = filled_markers[self.umarks.index(lab)]
				ax.scatter(points[:,0],points[:,1], points[:,2], c=mcolors[self.umarks.index(lab)],marker=curmarker)
			ax.set_xlabel('X Label')
			ax.set_ylabel('Y Label')
			ax.set_zlabel('Z Label')
			plt.show()

			fig1 = plt.figure()
			#ax = fig.add_subplot(111, projection='3d')
			ax1 = Axes3D(fig1)
			ax1.plot_trisurf(self.cellpat.data[:,0],self.cellpat.data[:,1],self.cellpat.data[:,2], color='y')
			plt.show()

class QuadScheme:
	"""
	Quadrature scheme for Berman Turner

	dimension : dimension of quad points (int)
	points : array of all point coordinates, real followed by dummy (nparray)
	ptlabs : array of point identities, 0 if dummy, 1 if real (nparray)
	quadweights : array of quadrature weights corresponding to each point in both the pattern and quadscheme, the first npoints are real, all following are dummy points (nparray)
	responses : array of calculated responses, equal to 1/quadweight for each real point, 0 for each dummy point (nparray)
	ptsonedge : array of points that happen to fall on the edge of a quad tile (unlikely), nparray
	nrp : number of real data points (int)
	numdummy : number of dummy points per dimension (int)
	pat : point pattern that the quadscheme was generated from/for (PointPattern)

	"""
	
	## initialize the quadrature scheme
		# numdummy : desired number of points on any given dimension (actually the cube (or square) root of the desired total number of dummy points)
		# creates a quadrature scheme based on equal sized tiling
	def __init__(self , pat , numdummy):

		## store the quad dimension
		self.dimension = pat.dimension

		## calculate the tile sizes
		tilesizex = (pat.box[0,1]-pat.box[0,0])/numdummy
		tilesizey = (pat.box[1,1]-pat.box[1,0])/numdummy

		## determine the dummy point coordinates by spacing evenly over each dimension
		xvals = np.linspace(pat.box[0,0]+tilesizex/2,pat.box[0,1]-tilesizex/2,numdummy)
		yvals = np.linspace(pat.box[1,0]+tilesizey/2,pat.box[1,1]-tilesizey/2,numdummy)
		
		## if 3D pattern, calculated z dimension for points
		if pat.dimension == 3:
			tilesizez = (pat.box[2,1]-pat.box[2,0])/numdummy

			zvals = np.linspace(pat.box[2,0]+tilesizez/2,pat.box[2,1]-tilesizez/2,numdummy)

			## manipulate the coordinate arrays to match point pattern formatting
			xv, yv, zv = np.meshgrid(xvals, yvals, zvals)
			xv = np.ndarray.flatten(xv)
			yv = np.ndarray.flatten(yv)
			zv = np.ndarray.flatten(zv)

			dummypts_all = np.column_stack((xv,yv,zv))

		else:
			## manipulate the coordinate arrays to match point pattern formatting
			xv, yv = np.meshgrid(xvals, yvals)
			xv = np.ndarray.flatten(xv)
			yv = np.ndarray.flatten(yv)
			dummypts_all = np.column_stack((xv,yv))

		c_hull = sps.ConvexHull(pat.cellpat.data)
		n_hull = sps.ConvexHull(pat.nucpat.data)
		c_region = sps.Delaunay(pat.cellpat.data[c_hull.vertices])
		n_region = sps.Delaunay(pat.nucpat.data[n_hull.vertices])
		
		## sort the dummy points for faster calculation
		rdps = []
		dummypts_all = dummypts_all[dummypts_all[:,0].argsort()]
		for pind in xrange(0,np.shape(dummypts_all)[0]):
			if in_hull(dummypts_all[pind,:],c_region) and not in_hull(dummypts_all[pind,:],n_region):
				rdps.append(dummypts_all[pind,:])
		dummypts = np.array(rdps)


		# if pat.dimension == 3:
		# 	## tilecounts holds the number of points per tile, there will be a total of numdummy^(ndim) tiles
		# 	tilecounts = np.zeros([numdummy*numdummy*numdummy,1])
		# else:
		# 	tilecounts = np.zeros([numdummy*numdummy,1])

		## pt2tile dict to store which point is in which tile
		pt2tile = {}
		tilecounter = 0

		## calculate the number of points in each tile for each point in both dummy and observed pattern
		if pat.dimension == 3:
			## set quadweights and quad responses
			self.quadweights = np.zeros([np.shape(dummypts)[0]+pat.npoints,1])
			self.responses = np.zeros([np.shape(dummypts)[0]+pat.npoints,1])
			tilecounts = np.zeros(np.shape(dummypts)[0])
			pt2tile = {}
			tilecounter = 0
			rps = sps.cKDTree(pat.points)

			for dpind in xrange(0,np.shape(dummypts)[0]):
				curdp = dummypts[dpind,:]
				tx = (curdp[0]-tilesizex/2,curdp[0]+tilesizex/2)
				ty = (curdp[1]-tilesizey/2,curdp[1]+tilesizey/2)
				tz = (curdp[2]-tilesizez/2,curdp[2]+tilesizez/2)
				cand_intile = rps.query_ball_point(curdp,round(max(tilesizex,tilesizey,tilesizez)))
				in_interval = lambda x,txl,tyl,tzl: True if txl[0] < x[0] < txl[1] and tyl[0] < x[1] < tyl[1] and tzl[0] < x[2] < tzl[1] else False
				intile = [pt for pt in cand_intile if in_interval(pat.points[pt,:],tx,ty,tz)]
				for val in intile:
					pt2tile[val] = tilecounter
				tilecounts[tilecounter] = len(intile)+1
				tilecounter += 1
		else:
			self.quadweights = np.zeros([np.shape(dummypts)[0]+pat.npoints,1])
			self.responses = np.zeros([np.shape(dummypts)[0]+pat.npoints,1])
			tilecounts = np.zeros(np.shape(dummypts)[0])
			pt2tile = {}
			tilecounter = 0
			rps = sps.cKDTree(pat.points)

			for dpind in xrange(0,np.shape(dummypts)[0]):
				curdp = dummypts[dpind,:]
				tx = (curdp[0]-tilesizex/2,curdp[0]+tilesizex/2)
				ty = (curdp[1]-tilesizey/2,curdp[1]+tilesizey/2)
				cand_intile = rps.query_ball_point(curdp,round(max(tilesizex,tilesizey)))
				in_interval = lambda x,txl,tyl: True if txl[0] < x[0] < txl[1] and tyl[0] < x[1] < tyl[1] else False
				intile = [pt for pt in cand_intile if in_interval(pat.points[pt,:],tx,ty)]
				for val in intile:
					pt2tile[val] = tilecounter
				tilecounts[tilecounter] = len(intile)+1
				tilecounter += 1
			# for curtilex in xrange(1,numdummy+1):
			# 	for curtiley in xrange(1,numdummy+1):
						
			# 			tx = ((curtilex-1)*tilesizex, curtilex*tilesizex)
			# 			ty = ((curtiley-1)*tilesizey, curtiley*tilesizey)
						

						

			# 			for rp in xrange(0,pat.npoints):
			# 				curpt = pat.points[rp,:]
			# 				if curpt[0] > tx[1]:
			# 					break
			# 				elif curpt[0] > tx[1] or curpt[0] < tx[0] or curpt[1] > ty[1] or curpt[1] < ty[0]:
			# 					continue
			# 				else:
			# 					pt2tile[rp] = tilecounter
			# 					tilecounts[tilecounter] += 1
								
						
			# 			for dp in xrange(0,numdummy*numdummy):
			# 				curpt = dummypts[dp,:]
			# 				if curpt[0] > tx[1]:
			# 					break
			# 				elif curpt[0] > tx[1] or curpt[0] < tx[0] or curpt[1] > ty[1] or curpt[1] < ty[0]:
			# 					continue
			# 				else:
								
			# 					pt2tile[pat.npoints+dp] = tilecounter
			# 					if pat.multimarked:
			# 						tilecounts[tilecounter] += len(pat.umarks)
			# 					else:
			# 						tilecounts[tilecounter] +=1
							
			# 			tilecounter = tilecounter + 1
		
			# self.quadweights = np.zeros([numdummy*numdummy+pat.npoints,1])
			# self.responses = np.zeros([numdummy*numdummy+pat.npoints,1])

		## if any points fall exactly on a tile edge (unlikely), record them
		ptsonedge = []
		
		## calculcate quadweights (1/number of points in tile) and quadresponses (0 if dummy, 1/weight if observed)
		for pt in xrange(0,pat.npoints):
			if pt in pt2tile:
				curtile = pt2tile[pt]
				if self.dimension == 3:
					self.quadweights[pt] = (tilesizex*tilesizey*tilesizez)/tilecounts[curtile]
					self.responses[pt] = 1/self.quadweights[pt]
				else:
					self.quadweights[pt] = (tilesizex*tilesizey)/tilecounts[curtile]
					self.responses[pt] = 1/self.quadweights[pt]
			else:
				ptsonedge.append(pt)

		for dpt in xrange(0,np.shape(dummypts)[0]):
			if self.dimension == 3:
				self.quadweights[dpt+pat.npoints] = (tilesizex*tilesizey*tilesizez)/tilecounts[dpt]
			else:
				self.quadweights[dpt+pat.npoints] = (tilesizex*tilesizey)/tilecounts[dpt]
			self.responses[dpt+pat.npoints] = 0

		for ind in xrange(0,pat.npoints+np.shape(dummypts)[0]):
			if self.quadweights[ind] == 0.0:
				if self.dimension == 3:
					self.quadweights[ind] = (tilesizex*tilesizey*tilesizez)/2
				else:
					self.quadweights[ind] = (tilesizex*tilesizey)/2
				self.responses[ind] = 1/self.quadweights[ind]

		## check if the pattern is multimarked
		if pat.multimarked:
			num_marks = len(pat.umarks)
			dum_pts = dummypts
			dum_marks = np.tile(pat.umarks[0],np.shape(dummypts)[0])
			self.marks = pat.marks
			dum_wts = self.quadweights[pat.npoints:]
			dum_resp = self.responses[pat.npoints:]
			self.umarks = pat.umarks
			self.multimarked = True
			for i in xrange(0,num_marks):
				if i==0:
					self.marks = np.concatenate((self.marks,dum_marks))
					self.points = np.concatenate((pat.points,dummypts))
				else:
					self.marks = np.concatenate((self.marks,np.tile(pat.umarks[i],np.shape(dummypts)[0])))
					self.points = np.concatenate((self.points,dummypts))
					self.quadweights = np.concatenate((self.quadweights,dum_wts))
					self.responses = np.concatenate((self.responses,dum_resp))


		else:

			## do some clean up to ensure that everything is in the right format
			self.points = np.concatenate((pat.points,dummypts))
			self.ptlabs = np.concatenate((np.ones(pat.npoints), np.zeros(np.shape(dummypts)[0])),0)
			self.multimarked = False

		self.ptsonedge = ptsonedge

		self.nrp = pat.npoints
		
		
		self.numdummy = np.shape(self.points)[0]-self.nrp

		## store the pattern for future calculation
		self.pat = pat


					
	## print the entire quadscheme
	def printQuad(self):

		print "QuadScheme"

		if self.dimension == 3:
			print "X  Y  Z  R/D  WT  RESP"
			for pt in xrange(0,np.shape(self.points)[0]):
				curpt = self.points[pt]
				curwt = self.quadweights[pt]
				curresp = self.responses[pt]
				curlab = self.ptlabs[pt]

				print curpt[0],curpt[1],curpt[2],curlab,curwt,curresp
		else:
			print "X  Y  R/D  WT  RESP"
			for pt in xrange(0,np.shape(self.points)[0]):
				curpt = self.points[pt]
				curwt = self.quadweights[pt]
				curresp = self.responses[pt]
				curlab = self.ptlabs[pt]

				print curpt[0],curpt[1],curlab,curwt,curresp
		print "Points on Quad edge: "
		for i in xrange(0,len(self.ptsonedge)):
			print self.ptsonedge[i]

	## print a relevant summary of the quadscheme
	def printQuadSummary(self):
		sys.stdout.write("Quadrature scheme with "+str(self.nrp)+" data points, "+str(self.numdummy)+" dummy points.\n")

	## plot the quadscheme using matplotlib
	def plotQuad(self):
		filled_markers = ('o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd')
		mcolors = ('b','g','r','c','m','y','k','w')

		if self.dimension == 3:
			
			fig = plt.figure()
				#ax = fig.add_subplot(111, projection='3d')
			ax = Axes3D(fig)
			for i in xrange(0,2):
				curmarker = filled_markers[i]
				if i == 0:
					ax.scatter(self.points[:self.nrp,0],self.points[:self.nrp,1], self.points[:self.nrp,2],c=mcolors[i],marker=curmarker)
				else:
					ax.scatter(self.points[self.nrp:,0],self.points[self.nrp:,1], self.points[self.nrp:,2],c=mcolors[i],marker=curmarker)
			ax.set_xlabel('X Label')
			ax.set_ylabel('Y Label')
			ax.set_zlabel('Z Label')
			plt.show()
		else:
				#fig = plt.figure()
				#ax = fig.subplot()
			fig,ax = plt.subplots(1,1)
			for i in xrange(0,2):
				curmarker = filled_markers[i]
				if i == 0:
					ax.scatter(self.points[:self.nrp,0],self.points[:self.nrp,1],c=mcolors[i],marker=curmarker)
				else:
					ax.scatter(self.points[self.nrp:,0],self.points[self.nrp:,1],c=mcolors[i],marker=curmarker)
			ax.set_xlabel('X Label')
			ax.set_ylabel('Y Label')
			xmax, ymax = self.points.max(axis=0)
			xmin, ymin = self.points.min(axis=0)
			ax.set_xlim(xmin,xmax)
			ax.set_ylim(ymin,ymax)
			ax.set_xticklabels([])
			ax.set_yticklabels([])
			plt.show()
			
class Interact:
	"""
	**** STILL UNDER CONSTRUCTION, NOT COMPLETE AND WORKAROUNDS ARE IMPLEMENTED FOR INTERACTION MODELS ****
	holds interaction types
	int_type : type of interaction model (strauss, multistrauss) (string)
	irreg_params : radii of interaction specified as a list 
		ex: 2 marks (1,2): [r12]
		ex: 4 marks (1,2,3,4) [r12 r13 r14 r23 r24 r34]

	"""
	def __init__(self, int_type, irreg_params={}):
		if int_type.lower() == "strauss":
			self.int_type = "strauss"
			self.radius = np.array(irreg_params)
		elif int_type.lower() == "multistrauss":
			self.int_type = "multistrauss"
			self.radius = irreg_params
		else:
			self.int_type = "poisson"
			self.radius = []
			
		
#############################################################
# Model fitting:

## main model fitting function, uses weighted general linear models to optimize parameters
	# uses beta version of weighted glm from statsmodels
def FitModel( quad , covar=DEFAULT , interaction=DEFAULT ):
	
	mod_glm = sm.GLM(quad.responses[:,0],sm.tools.tools.add_constant(covar),family=sm.families.Poisson(sm.families.links.log), freq_weights = quad.quadweights[:,0])
	# mod_glm = sm.GLM(quad.responses[:,0],covar,family=sm.families.Poisson(sm.families.links.log), freq_weights = quad.quadweights[:,0])
	res_glm = mod_glm.fit(disp=0)
	# betas = np.exp(res_glm.params)
	betas = res_glm.params
	loglike = res_glm.llf
	return betas,loglike

#############################################################
# cv likelihood tests
def inhomoLogLike(pat,coefs,cov_funcs):
	num_rand = 70000
	# generate a uniform sample from the cell region
	c_hull = sps.ConvexHull(pat.cellpat.data)
	n_hull = sps.ConvexHull(pat.nucpat.data)
	c_region = sps.Delaunay(pat.cellpat.data[c_hull.vertices])
	n_region = sps.Delaunay(pat.nucpat.data[n_hull.vertices])
	# rp = genRandPP(pat,num_rand,pat.box,c_region,n_region)
	rp = genRandPPBox(pat,num_rand,pat.box,c_region,n_region)
	rp.otherpat = pat.otherpat
	
	# calculate the covariate*coefficient value at all points
	trends = trendValue(rp,coefs,cov_funcs)
	print trends
	
	# norm = rp.cellarea*np.mean(trends)
	norm = winArea(rp.box)*np.mean(trends)
	print 'cell_area '+str(rp.cellarea)
	print norm
	# logdens = sum(np.log(trendValue(pat,coefs,cov_funcs)/(norm-np.log(0.01))))
	# logdens = sum(np.log(trendValue(pat,coefs,cov_funcs)/norm))
	# logdens = np.log(sum(trendValue(pat,coefs,cov_funcs)/norm))

	logdens = sum(np.log(trendValue(pat,coefs,cov_funcs)/norm))
	# logdens = sum(np.log(trendValue(pat,coefs,cov_funcs)))-pat.npoints*np.log(norm)
	# print logdens
	# print logdens+pat.npoints*np.log(.03)
	# print 'log_log_density '+str(np.log(logdens/pat.npoints) + np.log(0.01))
	# return np.log(logdens)/pat.npoints + np.log(0.05)
	# return logdens/pat.npoints + np.log(pat.cellarea/np.power(num_rand,1.0/3))
	# return (logdens/pat.npoints + np.log(winArea(rp.box)/np.power(num_rand,1.0/3)))/10
	return (logdens/pat.npoints + np.log(.001))
	# return logdens/pat.npoints + np.log(.01)
	# return logdens/pat.npoints
	


def trendValue(pat,coefs,cov_funcs):
	covars = []
	covars.append(np.ones([np.shape(pat.points)[0],1]))
	for func in cov_funcs:
		# cvr = func(pat)
		cvr = np.absolute(func(pat))
		if len(np.shape(cvr)) > 1:
			for col in xrange(0,np.shape(cvr)[1]):
				covars.append(cvr[:,col])
		else:
			covars.append(func(pat))
	# cdists = cellDist(pat)
	# ndists = nucDist(pat)
	# return np.exp([coefs[0]+cdists[ind]*coefs[1]+ndists[ind]*coefs[2] for ind in xrange(0,pat.npoints)])
	
	# return np.exp([sum((coefs[i]*covars[i][pind] for i in xrange(0,len(coefs)))) for pind in xrange(0,len(covars[0]))])
	
	return np.exp([sum((coefs[i]*covars[i][pind] for i in xrange(0,len(coefs)))) for pind in xrange(0,len(covars[0]))])
	# return [sum((coefs[i]*covars[i][pind] for i in xrange(0,len(coefs)))) for pind in xrange(0,len(covars[0]))]
	# return np.log([sum((coefs[i]*covars[i][pind] for i in xrange(0,len(coefs)))) for pind in xrange(0,len(covars[0]))])

def straussLpnorm(pat,coef,rad):
	niter = 200
	gam = np.exp(coef[0])
	# inhn = inhnorm(pat,coef)
	inter = np.zeros((niter,1))
	# print winArea(pat.box)
	beta = pat.npoints/winArea(pat.box)
	# print beta
	for k in xrange(0,niter):
		# rp = genHomoPP(pat.box,coef[0])
		rp = genHomoPP(pat.box,beta)
		randpat = PointPattern(rp,[],[],[],[],[])
		# c_inter = (sum(interactDist(randpat,rad))/2.0)*np.log(coef[0])
		c_inter = (sum(interactDist(randpat,rad))/2.0)*np.log(gam)
		# print c_inter
		inter[k] = c_inter
	norm = np.mean(inter)
	# print norm
	# print sum(interactDist(pat,rad)/2.0)*coef[0]
	# logden = (sum(interactDist(pat,rad))/2.0)*np.log(coef[0])-norm+1.0/(np.log(winArea(pat.box))*pat.npoints)
	logden = (sum(interactDist(pat,rad))/2.0)*np.log(gam)-norm+1.0/(np.log(winArea(pat.box))*pat.npoints)
	return logden

#############################################################
# p-value calculation

def MCTestStat(pat,dist_max,num_sim):
	real_K = kest(pat,dist_max)
	poiss_K = k_poiss(dist_max)
	L_real = np.mean(np.square((real_K - poiss_K)))
	L_avgs = np.zeros((num_sim,1))

	c_hull = sps.ConvexHull(pat.cellpat.data)
	n_hull = sps.ConvexHull(pat.nucpat.data)
	# region = sps.Delaunay(pat.cellpat)
	c_region = sps.Delaunay(pat.cellpat.data[c_hull.vertices])
	n_region = sps.Delaunay(pat.nucpat.data[n_hull.vertices])

	for nsim in xrange(0,num_sim):
		newpat = genRandPP(pat,pat.npoints,pat.box,c_region,n_region)
		# newpat.cellarea = pat.cellarea
		k_sim = kest(newpat,dist_max)
		Lval = np.square((k_sim - poiss_K))
		L_avgs[nsim] = np.mean(Lval)

	return L_real,L_avgs

def PValue(r_ts,sim_ts):
	return (1 + sum([1 for x in sim_ts if x >= r_ts]))/float(len(sim_ts))	

#############################################################
# utility functions

def numNeighbors(obj,rad):
	if isinstance(obj, PointPattern) or isinstance(obj, QuadScheme):
		points = obj.points
		
		if obj.multimarked == True:
			numMarks = len(obj.umarks)
			if numMarks != np.shape(rad)[0]:
				print "Interaction radius matrix must be of size [(number of marks) x (number of marks)], exiting"
				sys.exit()
			elif not (rad.transpose() == rad).all():
				print "Radius matrix must be symmetric, exiting."
				sys.exit()
			elif min(rad.flatten())<=0:
				print "Interaction radii must be non-negative, exiting."
				sys.exit()
			else:
				num_neighbors = {}
				pats = {}

				patind = {}
				for i in xrange(0,len(obj.umarks)):
					patind[obj.umarks[i]] = i 

				for mark in obj.umarks:
					pats[mark] = sps.cKDTree(obj.points[obj.marks==mark])

				for curpat in pats.keys():
					for otherpat in pats.keys():
						if curpat == otherpat:
							neighbor_inds = pats[curpat].query_ball_tree(pats[otherpat],rad[patind[curpat],patind[otherpat]])
							num_neighbors[(curpat,otherpat)] = [len(neigh)-1 for neigh in neighbor_inds]
						else:
							neighbor_inds = pats[curpat].query_ball_tree(pats[otherpat],rad[patind[curpat],patind[otherpat]])
							num_neighbors[(curpat,otherpat)] = [len(neigh) for neigh in neighbor_inds]

		else:
			tree = sps.cKDTree(obj.points)
			neighbor_inds = tree.query_ball_point(obj.points,rad)
			num_neighbors = [len(neigh) for neigh in neighbor_inds]


	else:
		print "Unrecognized object"
		sys.exit()

	return num_neighbors

def patDist(obj):
	if isinstance(obj, PointPattern):
		if isinstance(obj.otherpat,list):
			rdists = np.zeros([np.shape(obj.points)[0],len(obj.otherpat)])
			for opind, op in enumerate(obj.otherpat):
				op_tree = sps.cKDTree(op)
				dists = op_tree.query(obj.points,2)[0]
				for ind,val in enumerate(dists):
					if val[0] != 0:
						rdists[ind,opind] = val[0]
					else:
						rdists[ind,opind] = val[1]
			return rdists
		else:
			op_tree = sps.cKDTree(obj.otherpat)
			dists = op_tree.query(obj.points,2)[0]
			rdists = np.zeros([np.shape(dists)[0]])
			for ind,val in enumerate(dists):
				if val[0] != 0:
					rdists[ind] = val[0]
				else:
					rdists[ind] = val[1]
			return rdists

	elif isinstance(obj, QuadScheme):
		if isinstance(obj.pat.otherpat,list):
			rdists = np.zeros([np.shape(obj.points)[0],len(obj.pat.otherpat)])
			for opind, op in enumerate(obj.pat.otherpat):
				op_tree = sps.cKDTree(op)
				dists = op_tree.query(obj.points,2)[0]
				for ind,val in enumerate(dists):
					if val[0] != 0:
						rdists[ind,opind] = val[0]
					else:
						rdists[ind,opind] = val[1]
			return rdists
		else:
			op_tree = sps.cKDTree(obj.pat.otherpat)
			dists = op_tree.query(obj.points,2)[0]
			rdists = np.zeros([np.shape(dists)[0]])
			for ind,val in enumerate(dists):
				if val[0] != 0:
					rdists[ind] = val[0]
				else:
					rdists[ind] = val[1]
			return rdists
		# return obj.pat.cellpat.query(obj.points,1)[0]
	else:
		print "Cannot calculate covariance, instance passed was not numpy array or PointPattern, exiting."
		sys.exit()
# NEW

def fracDist(obj):
	return cellDist(obj)/nucDist(obj)

def cellDist(obj):
	if isinstance(obj, PointPattern):
		if obj.cellpat == []:
			print "Cell pattern required for cellDist function, returning empty"
			return []
		else:
			dists = obj.cellpat.query(obj.points,2)[0]
			rdists = np.zeros([np.shape(dists)[0]])
			for ind,val in enumerate(dists):
				if val[0] != 0:
					rdists[ind] = val[0]
				else:
					rdists[ind] = val[1]
			return rdists

	elif isinstance(obj, QuadScheme):
		if obj.pat.cellpat == []:
			print "Cell pattern required for cellDist function, returning empty"
			return []
		else:
			dists = obj.pat.cellpat.query(obj.points,2)[0]
			rdists = np.zeros([np.shape(dists)[0]])
			for ind,val in enumerate(dists):
				if val[0] != 0:
					rdists[ind] = val[0]
				else:
					rdists[ind] = val[1]
			return rdists
			# return obj.pat.cellpat.query(obj.points,1)[0]
	else:
		print "Cannot calculate covariance, instance passed was not QuadScheme or PointPattern, exiting."
		sys.exit()
	# print landmark
# NEW
def nucDist(obj):
	if isinstance(obj, PointPattern):
		if obj.nucpat == []:
			print "Nuclear pattern required for cellDist function, returning empty"
			return []
		else:
			dists = obj.nucpat.query(obj.points,2)[0]
			rdists = np.zeros([np.shape(dists)[0]])
			for ind,val in enumerate(dists):
				if val[0] != 0:
					rdists[ind] = val[0]
				else:
					rdists[ind] = val[1]
			return rdists
			# return obj.nucpat.query(obj.points,1)[0]
	elif isinstance(obj, QuadScheme):
		if obj.pat.nucpat == []:
			print "Nuclear pattern required for cellDist function, returning empty"
			return []
		else:
			dists = obj.pat.nucpat.query(obj.points,2)[0]
			rdists = np.zeros([np.shape(dists)[0]])
			for ind,val in enumerate(dists):
				if val[0] != 0:
					rdists[ind] = val[0]
				else:
					rdists[ind] = val[1]
			return rdists
			# return obj.pat.nucpat.query(obj.points,1)[0]
	else:
		print "Cannot calculate covariance, instance passed was not QuadScheme or PointPattern, exiting."
		sys.exit()

# kest: calculates ripleys K function for radii in [0:dist_max]
## returns ripleys K for the point pattern 
def kest(pat,dist_max):
	tree = sps.cKDTree(pat.points)
	intensity = pat.cellarea/(pat.npoints*(pat.npoints-1))
	sum_func = lambda rad: intensity*sum([len(val)-1 for val in tree.query_ball_point(pat.points,rad)])
	
	# result = integrate.quad(sum_func,0,dists)

	dists = np.linspace(0,dist_max,500)

	K = np.zeros((500,1))
	for ind,val in enumerate(dists):
		K[ind] = sum_func(val)


	# poiss_K = integrate.quad(lambda t: np.pi*t**2, 0,dists)
	return K

def k_poiss(dist_max):
	poiss_func = lambda t: np.pi*t**2
	dists = np.linspace(0,dist_max,500)
	K_poiss = np.zeros((500,1))
	for ind,val in enumerate(dists):
		K_poiss[ind] = poiss_func(val)
	return K_poiss

def in_hull(p, region):
	"""
	Test if points in `p` are in `hull`

	`p` should be a `NxK` coordinates of `N` points in `K` dimensions
	`hull` is either a scipy.spatial.Delaunay object or the `MxK` array of the 
	coordinates of `M` points in `K`dimensions for which Delaunay triangulation
	will be computed
	"""
	if not isinstance(region,sps.Delaunay):
		region = sps.Delaunay(region)

	return region.find_simplex(p)>=0

def hullVol(hull):
	if np.shape(hull.points)[1] == 3:
		simplices = np.column_stack((np.repeat(hull.vertices[0], hull.nsimplex), hull.simplices))
		tets = hull.points[simplices]
		return np.sum(tetrahedron_volume(tets[:, 0], tets[:, 1], tets[:, 2], tets[:, 3]))
	else:
		lines = np.hstack([hull.points,np.roll(hull.points,-1,axis=0)])
		area = 0.5*abs(sum(x1*y2-x2*y1 for x1,y1,x2,y2 in lines))
		return area

def tetrahedron_volume(a, b, c, d):
	return np.abs(np.einsum('ij,ij->i', a-d, np.cross(b-d, c-d))) / 6

def winArea(x):
	if np.shape(x)[0] == 3:
		return (x[0,1]-x[0,0])*(x[1,1]-x[1,0])*(x[2,1]-x[2,0])
	else:
		return (x[0,1]-x[0,0])*(x[1,1]-x[1,0])

#############################################################
# synthesize pats

def genRandPP(pat,n,win,c_region,n_region):
	# win is an np.array where win=[xmin xmax; ymin ymax; zmin zmax]

	if np.shape(win)[0] == 3:
		x=stats.uniform.rvs(win[0,0],win[0,1],n)
		y=stats.uniform.rvs(win[1,0],win[1,1],n)
		z=stats.uniform.rvs(win[2,0],win[2,1],n)
		pos = np.column_stack((x,y,z))
	else:
		x=stats.uniform.rvs(win[0,0],win[0,1],n)
		y=stats.uniform.rvs(win[1,0],win[1,1],n)
		pos = np.column_stack((x,y))

	for p in xrange(0,np.shape(pos)[0]):
		if in_hull(pos[p,:],c_region) and not in_hull(pos[p,:],n_region):
			continue
		else: 
			if np.shape(win)[0] == 3:
				inflag = False
				while inflag is False:
					x=stats.uniform.rvs(win[0,0],win[0,1],1)
					y=stats.uniform.rvs(win[1,0],win[1,1],1)
					z=stats.uniform.rvs(win[2,0],win[2,1],1)

					if in_hull(np.column_stack((x,y,z)),c_region) and not in_hull(np.column_stack((x,y,z)),n_region):
						pos[p,0] = x
						pos[p,1] = y
						pos[p,2] = z
						inflag = True
			else:
				inflag = False
				while inflag is False:
					x=stats.uniform.rvs(win[0,0],win[0,1],1)
					y=stats.uniform.rvs(win[1,0],win[1,1],1)

					if in_hull(np.column_stack((x,y)),c_region) and not in_hull(np.column_stack((x,y)),n_region):
						pos[p,0] = x
						pos[p,1] = y
						inflag = True

	return PointPattern(pos,mark=pat.marks,area=pat.cellarea,cellpat=pat.cellpat.data,nucpat=pat.nucpat.data,dimension=pat.dimension,box=pat.box)

def genRandPPBox(pat,n,win,c_region,n_region):
	# win is an np.array where win=[xmin xmax; ymin ymax; zmin zmax]

	if np.shape(win)[0] == 3:
		x=stats.uniform.rvs(win[0,0],win[0,1],n)
		y=stats.uniform.rvs(win[1,0],win[1,1],n)
		z=stats.uniform.rvs(win[2,0],win[2,1],n)
		pos = np.column_stack((x,y,z))
	else:
		x=stats.uniform.rvs(win[0,0],win[0,1],n)
		y=stats.uniform.rvs(win[1,0],win[1,1],n)
		pos = np.column_stack((x,y))

	return PointPattern(pos,mark=pat.marks,area=pat.cellarea,cellpat=pat.cellpat.data,nucpat=pat.nucpat.data,dimension=pat.dimension,box=pat.box)

def GenPoissonPP(intensity,xl,yl,zl,marks=[]):
	'''
	Determines the number of events `N` for a rectangular region,
	given the rate `rt` and the dimensions, `Dx`, `Dy`.
	Returns a <2xN> NumPy array.
	'''
	
	N = stats.poisson(intensity*xl*yl*zl).rvs()
	x = stats.uniform.rvs(0,xl,((N,1)))
	y = stats.uniform.rvs(0,yl,((N,1)))
	z = stats.uniform.rvs(0,zl,((N,1)))
	P = np.hstack((x,y,z))
	markind = ''
	if marks != []:
		markind = np.random.randint(0,high=len(marks),size=np.shape(P)[0])
	mark = []
	for val in markind:
		mark.append(marks[val])
	pat = PointPattern(P,mark)
	return pat

def GenPoisson2D(intensity,xl,yl,marks=[]):
	'''
	Determines the number of events `N` for a rectangular region,
	given the rate `rt` and the dimensions, `Dx`, `Dy`.
	Returns a <2xN> NumPy array.
	'''
	
	N = stats.poisson(intensity*xl*yl).rvs()
	x = stats.uniform.rvs(0,xl,((N,1)))
	y = stats.uniform.rvs(0,yl,((N,1)))
	P = np.hstack((x,y))
	markind = ''
	if marks != []:
		markind = np.random.randint(0,high=len(marks),size=np.shape(P)[0])
	mark = []
	for val in markind:
		mark.append(marks[val])
	pat = PointPattern(P,mark)
	return pat

def GenStraussPP(dim,gam,r,n,its,mark=""):

	x = np.random.rand(n,dim)
	l = np.zeros(its);
	for i in xrange(0,its):
		J = np.ceil(n*np.random.random_sample());
		if J >= n-1:
			J = n-1
		y = x
		y[J,:] = y[J,:] + 0.1*np.random.normal(loc = 0, scale = 1,size =dim)
		a = numpairs(y,r)  
		b = numpairs(x,r)
		if (np.amax(np.amax(y)) > 1 or np.amin(np.amin(y)) <0):
			alpha =0
		elif a < b:
			alpha =1
		else:
			alpha = np.power(gam,numpairs(y,r))/np.power(gam,numpairs(x,r))
		R = np.random.random_sample()< alpha
		x = R*y + (1-R)*x
		l[i] = numpairs(x,r)
		#plt.plot(x[:,0],x[:,1],'.')
	return PointPattern(x,mark)

def GenStraussMH(dim,gam,beta,r,n,its,mark=""):
	x = np.random.rand(n,dim)
	for i in xrange(0,its):
		n = np.shape(x)[0]
		B = np.random.random_sample() < 0.5
		if B:
			xnew = np.random.random(size=dim)
			y = np.vstack((x,xnew))
			n = n+1;
		else:
			a = np.ceil(n*np.random.random_sample())
			if a >= n-1:
				a = n-1
			b = x[a,:]
			idx = np.where(abs((x[:,np.newaxis,:] - b)).sum(axis=2)==0)
			y = x
			y = np.delete(y,idx[0],0)
		
		if np.amax(np.amax(y)) > 1 or np.amin(np.amin(y)) < 0:
			alpha =0;
		elif (numpairs(y,r) < numpairs(x,r)):
			alpha =1
		elif B:
			alpha = beta*np.power(gam,(numpairs(y,r) - numpairs(x,r)))/n
		else:
			alpha = n*np.power(gam,(numpairs(y,r) - numpairs(x,r)))/beta

		if (np.random.random_sample() < alpha):
			x = y;
		#plt.plot(x[:,0],x[:,1],'.');
	return PointPattern(x,mark)

def GenMultiStrauss(dim,gams,rs,ns,its,marks):
	if dim == '2D':
			x = np.random.rand(n,2)
			l = np.zeros(its);
			for i in xrange(0,its):
				J = np.ceil(n*np.random.random_sample());
				if J >= n-1:
					J = n-1
				y = x;
				y[J,:] = y[J,:] + 0.1*np.random.normal(loc = 0, scale = 1,size =2)
		
				a = numpairs(y,r)  
				b = numpairs(x,r)
				if (np.amax(np.amax(y)) > 1 or np.amin(np.amin(y)) <0):
					alpha =0
				elif a < b:
					alpha =1
				else:
					alpha = np.power(gam,numpairs(y,r))/np.power(gam,numpairs(x,r))
				R = np.random.random_sample()< alpha
				x = R*y + (1-R)*x
				l[i] = numpairs(x,r)
				#plt.plot(x[:,0],x[:,1],'.')
			return PointPattern(x,mark)

def simulateInhomogenous(pat,betas,cov_funcs,dists=[]):
	if dists == []:
		dists = np.ones([np.shape(pat.points)[0],len(cov_funcs)+1])
		for ind,func in enumerate(cov_funcs):
			cvr = np.absolute(func(pat))
			if len(np.shape(cvr))>1:
				cvr = cvr[:,0]
			dists[:,ind+1] = cvr
	
	# calculate L*
	get_lam = lambda dists,betas: np.dot(dists,betas)
	max_lam = max([get_lam(dists[i,:],betas) for i in xrange(0,pat.npoints)])


	# get number of points as random variable
	N = stats.poisson(max_lam/pat.cellarea).rvs()
	print N
	# N = 10000
	c_hull = sps.ConvexHull(pat.cellpat.data)
	n_hull = sps.ConvexHull(pat.nucpat.data)
	c_region = sps.Delaunay(pat.cellpat.data[c_hull.vertices])
	n_region = sps.Delaunay(pat.nucpat.data[n_hull.vertices])
	base_pat = genRandPP(pat,N,pat.box,c_region,n_region)
	# base_pat = genRandPPBox(pat,N,pat.box,c_region,n_region)
	base_pat.otherpat = pat.otherpat
	base_thinning_prob = trendValue(base_pat,np.log(betas),cov_funcs)/max_lam

	final_pt_inds = []
	for ind,v in enumerate(base_thinning_prob):
		if stats.uniform.rvs(0,1,1)	> v:
			final_pt_inds.append(ind)

	base_pat.points = base_pat.points[final_pt_inds,:]
	return base_pat
